//  pnpp.java - parallel name preprocessing
//
//  Licensed Materials - Property of IBM
//  "Restricted Materials of IBM"
//  5724-Q20
//   Copyright IBM Corp. 2010

import java.io.*;
import java.util.*;

public class pnpp
{
	//  This application reads a CSV file containing name
	//  records from the standard input, splits these records
	//  into multiple sub-lists and runs multiple parallel NPP
	//  processes to perform name analysis (the first phase of
	//  name preprocessing) on the name records, then combines
	//  individual results into a single list and runs another NPP
	//  process to perform unique name identification.
	public static void main(String[] args) throws Throwable
	{
		//  include phase 1, name analysis
		boolean phase1 = true;
		//  include phase 2, identification of unique names
		boolean phase2 = true;
		//  number of Analyzer processes
		int processes = 4;
		//  number of output sublists
		int sublists = 2;
		//  base name of configuration template files
		String template = "pnpp";
		//  temporary directory
		String tempDir = ".";
		//  input encoding
		String encoding = "UTF-8";
		//  output file name base
		String output = "pnpp";

		//  parse command line
		for ( String arg : args )
		{
			if ( arg.startsWith("-processes=") )
				processes = Integer.parseInt(arg.substring(11));
			else if ( arg.startsWith("-sublists=") )
				sublists = Integer.parseInt(arg.substring(10));
			else if ( arg.startsWith("-config=") )
				template = arg.substring(8);
			else if ( arg.startsWith("-tempdir=") )
				tempDir = arg.substring(9);
			else if ( arg.startsWith("-encoding=") )
				encoding = arg.substring(10);
			else if ( arg.startsWith("-output=") )
				output = arg.substring(8);
			else if ( arg.equals("-phase1") )
				phase1 = false;
			else if ( arg.equals("-phase2") )
				phase2 = false;
			else
			{
				System.err.println("Parameter error: " + arg);
				System.exit(1);
			}
		}

		//  perform name analysis ??
		if ( phase1 )
		{
			//  split input file into sections, one per process
			List<String> fileNames = new ArrayList<String>();
			List<PrintStream> sections = new ArrayList<PrintStream>();
			for ( int i = 0; i < processes; ++i )
			{
				String fileName = tempDir + "/section" + i;
				fileNames.add(fileName);
				File file = new File(fileName);
				file.deleteOnExit();
				sections.add(new PrintStream(new FileOutputStream(file), true, "UTF-8"));
			}
			int current = 0;
			BufferedReader in = new BufferedReader(new InputStreamReader(System.in, encoding));
			String line;
			while ( (line = in.readLine()) != null )
			{
				if ( current >= sections.size() )
					current = 0;
				sections.get(current++).println(line);
			}
			//  close all section files, so they can be automatically deleted
			for ( PrintStream ps : sections )
				ps.close();

			//  start multiple analyzer processes
			ConfigFile.setTempDir(new File(tempDir));
			List<Analyzer> procs = new ArrayList<Analyzer>();
			for ( String fn : fileNames )
				procs.add(new Analyzer(template, fn));

			//  wait for analyzers to complete, check for failures
			boolean analyzerError = false;
			for ( Analyzer an : procs )
				if ( an.waitFor() != 0 )
				{
					System.err.println("Analyzer failure");
					analyzerError = true;
				}

			//  scan for processing errors
			for ( String fn : fileNames )
			{
				File file = new File(fn + ".error");
				file.deleteOnExit();
				BufferedReader err = new BufferedReader(
				                         new InputStreamReader(
				                             new FileInputStream(file) ) );
				String errLine;
				while ( (errLine = err.readLine()) != null )
				{
					if ( errLine.indexOf("discarded") != -1 )
					{
						//  report errors
						System.err.println("Analysis error: " + errLine);
						analyzerError = true;
					}
				}
				//  force close to allow automatic deletion
				err.close();
			}

			//  abort now on any analysis error
			if ( analyzerError )
				System.exit(1);

			//  concatenate analysis output to a single file
			SequenceInputStream sis = new SequenceInputStream(new InterimFileEnumerator(fileNames));
			FileOutputStream interim = new FileOutputStream("./" + output + ".npp");
			byte[] buffer = new byte[128 * 1024];
			int nr;
			while ( (nr = sis.read(buffer, 0, buffer.length)) != -1 )
				interim.write(buffer, 0, nr);
			sis.close();
			interim.close();
		}

		//  skip phase 2 ??
		if ( !phase2 )
			System.exit(0);

		//  create config file for unique name identification
		ConfigFile unique = new ConfigFile(template + ".2.config", output, Integer.toString(sublists));

		//  run unique name identification
		Process proc = Runtime.getRuntime().exec("npp " + unique);
		int exitCode = proc.waitFor();
		if ( exitCode != 0 )
			System.err.println("Error exit: " + exitCode);
		System.exit(exitCode);
	}
}

//  a configuration file
//  Creates a copy of a template configuration file, substituting
//  variables for specific text sequences.
class ConfigFile
{
	private static File tempDir = new File(".");
	private String name;

	public static void setTempDir(File dir)
		{ tempDir = dir; }

	public ConfigFile(String template, String inputName, String sublists) throws Throwable
	{
		//  create temporary file
		File cfg = File.createTempFile("pnpp", ".config", tempDir);
		cfg.deleteOnExit();
		//  save file name
		name = cfg.getAbsolutePath();

		//  copy template file
		//  replace all "%INPUT%" sequences with input file name
		//  replace all "%SUBLISTS%" sequences with number of sublists
		BufferedReader in = new BufferedReader(
		                        new InputStreamReader(
		                            new FileInputStream(template), "UTF-8" ) );
		PrintStream out = new PrintStream(new FileOutputStream(cfg), true, "UTF-8");
		String line;
		while ( (line = in.readLine()) != null )
		{
			line = line.replace("%INPUT%", inputName);
			line = line.replace("%SUBLISTS%", sublists);
			out.println(line);
		}

		//  must release file in order to delete later
		out.close();
	}

	public String toString()
		{ return name; }
}

//  an analysis process
//  Manages analysis of a single sub-list by running a 'npp' process.
class Analyzer
{
	private Process proc;
	private ConfigFile config;

	public Analyzer(String template, String input) throws Throwable
	{
		//  create config file
		config = new ConfigFile(template + ".1.config", input, "");

		//  run name analysis
		proc = Runtime.getRuntime().exec("npp " + config);
	}

	public int waitFor() throws Throwable
		{ return proc.waitFor(); }
}

//  enumerate files for concatenation
//  Enumerates a list of intermediate files for concatenation, 
//  ensuring that the intermediate files are deleted after use.
class InterimFileEnumerator implements Enumeration<InputStream>
{
	private List<String> fileNames;
	private int current = 0;

	public InterimFileEnumerator(List<String> fileNames)
		{ this.fileNames = fileNames; }

	public boolean hasMoreElements()
		{ return current < fileNames.size(); }

	public InputStream nextElement()
	{
		if ( !hasMoreElements() )
			throw new NoSuchElementException("No more files.");

		String name = fileNames.get(current++) + ".npp";
		File file = new File(name);
		file.deleteOnExit();
		InputStream stream = null;
		try
		{
			stream = new FileInputStream(file);
		}
		catch ( FileNotFoundException e )
		{
			System.err.println("File not found: " + name);
		}

		return stream;
	}
}
